<?php
// $Id: EstimateDates.inc,v 1.1 2006/05/01 00:59:21 amnonj Exp $
// EstimateDates.inc By Amnon Jonas 23-Apr-2006

function family_auto_privacy() {
  $content = "<p>This will run an algorithm to automatically set privacy<br>".
              "Generally, the algorithm sets living people and their generation as private. ".
              "The algorithm tries to evaluate the generation of people, using their own data and data of their relatives. ".
              "To protect privacy, errors are usually biased for protecting data.</p>".
              "<p>Go to Administer Access control to set which roles may access public or private records</p>".
              "<p>The results cannot be used to determine if a specific individual is actually living. ".
              "If you have sensitive data, please check manually that the data is set as private. ".
              "You should be aware that any database may be hacked and there is no guarantee for your data security</p>".
              "<p>For large files this may take a few minutes.</p>".
              "<p><a href=\"/admin/family/privacy_done\">Start auto privacy</a></p>";
  return $content;
}

/**
 * Decode GEDCOM formatted date field and store it as a range array
 * Currently support only Gregorian dates for years 100-9999
 */
function family_decode_date($date) {
  $range = array( 'MIN' => -99999,'MAX' => 99999,'AVG' => 2000,'VAR'  => 9e9); //Default values
  if (preg_match("/^BEF.*?(\d{3,4})/i", $date , $match)) {
    $range['MAX']=$match[1];
    return $range;
  }
  if (preg_match("/^AFT.*?(\d{3,4})/i", $date , $match)) {
    $range['MIN']=$match[1];
    return $range;
  }
  if (preg_match("/^BET.*?(\d{3,4}).*AND.*?(\d{3,4})/i", $date , $match)) {
    $range['MIN']=$match[1];
    $range['MAX']=$match[2];
    return $range;
  }
  if (preg_match("/^ABT.*?(\d{3,4})/i", $date , $match)) {
    $range['AVG']=$match[1];
    $range['VAR']=25;
    return $range;
  }
  if (preg_match("/^.*?(\d{3,4}).*?$/i", $date , $match)) {
    $range['AVG']=$match[1];
    $range['VAR']=0;
    return $range;
  }
  return $range;
}

/**
 * Fix range array. Take care of min>max and use information from (min,max) to update (avg,var)
 * and vice versa
 */
function family_fix_range($range) {
  if ($range['VAR']==0){
    $range['MIN']=$range['AVG'];
    $range['MAX']=$range['AVG'];
  }
  if ($range['MIN'] > $range['MAX']) {
    $range['VAR']=0;
    $range['AVG']=($range['MIN'] + $range['MAX'])/2;
    $range['MAX']=$range['AVG'];
    $range['MIN']=$range['AVG'];
  }
  else {
    if (pow(($range['MAX'] - $range['MIN'])/2,2)<$range['VAR']) {
      $range['AVG']=($range['MIN'] + $range['MAX'])/2;
      $range['VAR']=pow(($range['MAX'] - $range['MIN'])/2,2);
    }
  }
  return $range;
}

/**
 * Add two range arrays
 * Used to add offset range between two events to get range of one events, based on the other event
 */
function family_add_ranges($range1,$range2) {
  $range['MIN']=$range1['MIN']+$range2['MIN'];
  $range['MAX']=$range1['MAX']+$range2['MAX'];
  $range['AVG']=$range1['AVG']+$range2['AVG'];
  $range['VAR']=$range1['VAR']+$range2['VAR'];
  $range = family_fix_range($range);
  return $range;
}

/**
 * Merge two range arrays that refer to the same event. It selectss the more accurate data of the two.
 */
function family_merge_ranges($range1,$range2) {
  $range['MIN']=max($range1['MIN'],$range2['MIN']);
  $range['MAX']=min($range1['MAX'],$range2['MAX']);
  if ($range1['VAR']==$range2['VAR']) {
    $range['AVG']=($range1['AVG']+$range2['AVG'])/2;
    $range['VAR']=$range1['VAR'];
  }
  else {
    if ($range1['VAR'] < $range2['VAR']) {
      $range['AVG']=$range1['AVG'];
      $range['VAR']=$range1['VAR'];
    }
    else {
      $range['AVG']=$range2['AVG'];
      $range['VAR']=$range2['VAR'];
    }
  }
  $range = family_fix_range($range);
  return $range;
}

/**
 * Run an estimation algorithm that estimate dates ranges for birth and death events.
 * This is used to determine which records should be private
 */
function family_auto_privacy_done()
{
  $birth2death = array('MIN' => 0, 'MAX' => 100,'AVG' => 60, 'VAR' => 1600);
  $death2birth = array('MIN' => -100,'MAX' => 0,'AVG' => -60, 'VAR' => 1600);
  $parentbirth2childbirth = array('MIN' => 16, 'MAX' => 56, 'AVG' => 32, 'VAR' => 256);
  $parentdeath2childbirth = array('MIN' => -75, 'MAX' => 0, 'AVG' => -30, 'VAR' => 900);
  $childbirth2parentbirth = array('MIN' => -56, 'MAX' => -16, 'AVG' => -32, 'VAR' => 256);
  $childbirth2parentdeath = array('MIN' => 0, 'MAX' => 75, 'AVG' => 30, 'VAR' => 900);
  $today = getdate();
  $thisyear=$today['year'];

//Save the individuals and children queries results to save doing the same query each iteration
  $individuals = db_query("SELECT fid FROM {family_facts} WHERE fact_code = 'INDI'");

// Collect Individuals Birth and Death data
  if ($individuals && (db_num_rows($individuals)>0)) {
    for ($i = 0; $i < db_num_rows($individuals); $i++) {
      $fid=db_result($individuals,$i);
      if ($birt_fid = family_get_subfid_by_code("BIRT", $fid)){
        $birthrange[$fid] = family_fix_range(family_decode_date(family_get_subvalue_by_code("DATE", $birt_fid)));
        $birthrange[$fid]['MAX'] = min($birthrange[$fid]['MAX'], $thisyear);
      }
      else {
        $birthrange[$fid] = array('MIN' => -99999, 'MAX' => $thisyear, 'AVG' => 2000, 'VAR' => 9e9); //Default values
      }
      if ($deat_fid = family_get_subfid_by_code("DEAT", $fid)){
        $deathrange[$fid] = family_fix_range(family_decode_date(family_get_subvalue_by_code("DATE", $deat_fid)));
      }
      else {
        $deathrange[$fid] = array('MIN' => -99999, 'MAX' => 99999, 'AVG' => 2000, 'VAR' => 9e9); //Default values
      }
    }
  }
  else {
    return;
  }

  $number_iterations = 8;
  for ($iterations = 0; $iterations < $number_iterations; $iterations++) {
//Compare Death and Birth
    echo "<!-- -->"; //Just send an empy HTML comment to avoid client timeout while waiting
    for ($i = 0; $i < db_num_rows($individuals); $i++) {
      $fid=db_result($individuals,$i);
      $deathrange[$fid] = family_merge_ranges($deathrange[$fid],family_add_ranges($birthrange[$fid],$birth2death));
      $birthrange[$fid] = family_merge_ranges($birthrange[$fid],family_add_ranges($deathrange[$fid],$death2birth));
    }
//Compare parents and children
    $children = db_query("SELECT fid1,fid2 FROM {family_relations} WHERE relation_description = 'CHIL'");
    while($child = db_fetch_array($children)) {
      $fid1=$child['fid1']; //Child fid
      $fam=$child['fid2']; //Family fid
      $parents = db_query("SELECT fid1 FROM {family_relations} WHERE (fid2 = %d AND (relation_description='%s' OR relation_description='%s'))",$fam,"HUSB","WIFE");
      while($parent = db_fetch_array($parents)) {
        $fid2=$parent['fid1']; //parent fid
        $birthrange[$fid1] = family_merge_ranges($birthrange[$fid1],family_add_ranges($birthrange[$fid2],$parentbirth2childbirth));
        $birthrange[$fid1] = family_merge_ranges($birthrange[$fid1],family_add_ranges($deathrange[$fid2],$parentdeath2childbirth));
        $birthrange[$fid2] = family_merge_ranges($birthrange[$fid2],family_add_ranges($birthrange[$fid1],$childbirth2parentbirth));
        $deathrange[$fid2] = family_merge_ranges($deathrange[$fid2],family_add_ranges($birthrange[$fid1],$childbirth2parentdeath));
      }
    }
  }

// Update privacy information
// Privacy is stored in fact _PRV (family module defined tag)
// 0 or none: Auto (public until estimation is run), 1: Always public (overide auto), 2: Always private (overide auto)
// 3: Public (set by estimation algorithm), 4: Private (set by estimation algorithm) 
  $PrivacyLevel = array( 0 => "Auto", 1 => "Public (Manually)", 2=>"Private (Manually)", 3 => "Public (Auto)", 4=>"Private (Auto)");
  for ($i = 0; $i < db_num_rows($individuals); $i++) {
    $fid=db_result($individuals,$i);
    if ($privacy_fid = family_get_subfid_by_code("_PRV", $fid)){
      $privacy = family_get_subvalue_by_code("_PRV", $fid);
    }
    else {
      $privacy_fid = family_insert_fact('_PRV', $privacy);
      family_insert_relation($privacy_fid, $fid, 'FACT');
      $privacy=0;
    }
     if (($privacy==0) || ($privacy==3) || ($privacy==4)) {
       if (($deathrange[$fid]['MAX']<=$thisyear) || ($deathrange[$fid]['AVG']+sqrt($deathrange[$fid]['VAR'])<=$thisyear)) {
         $privacy=3;
       }
       else {
         $privacy=4;
       }
       family_update_fact($privacy_fid, $privacy);
     }

//Write Estimated dates in Biography, for testing
//    $nid = db_result(db_query("SELECT nid FROM family_facts WHERE fid = %d", $fid));
//    $node=node_load($nid);
//    $content = "<strong>Estimated Dates (calculated by Family Module)</strong>";
//    $content.= "<ul>Birth:&nbsp;&nbsp;".$birthrange[$fid]['MIN']."- ".$birthrange[$fid]['MAX']."&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;".round($birthrange[$fid]['AVG'])."&plusmn;".round(sqrt($birthrange[$fid]['VAR']))."</ul>";
//    $content.= "<ul>Death: ".$deathrange[$fid]['MIN']."- ".$deathrange[$fid]['MAX']."&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;".round($deathrange[$fid]['AVG'])."&plusmn;".round(sqrt($deathrange[$fid]['VAR']))."</ul>";
//    $content.="<p>Privacy status: ".$PrivacyLevel[$privacy]."</p>";
//    $node->body = $content;
//    node_save(&$node);

  }
  return "<p>Done</p>";
}

